/*
 * Copyright (C) 1999-2020. Christian Heller.
 *
 * This file is part of the Cybernetics Oriented Interpreter (CYBOI).
 *
 * CYBOI is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * CYBOI is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CYBOI. If not, see <http://www.gnu.org/licenses/>.
 *
 * Cybernetics Oriented Programming (CYBOP) <http://www.cybop.org/>
 * CYBOP Developers <cybop-developers@nongnu.org>
 *
 * @version CYBOP 0.21.0 2020-07-29
 * @author Christian Heller <christian.heller@cybop.org>
 */

#ifndef SET_MODE_UNIX_TERMINAL_STARTER_SOURCE
#define SET_MODE_UNIX_TERMINAL_STARTER_SOURCE

#include <errno.h> // errno
#include <termios.h> // struct termios

#include "../../../../constant/model/cyboi/log/level_log_cyboi_model.c"
#include "../../../../constant/model/cyboi/state/integer_state_cyboi_model.c"
#include "../../../../constant/model/cyboi/state/pointer_state_cyboi_model.c"
#include "../../../../executor/copier/integer_copier.c"
#include "../../../../logger/logger.c"

/**
 * Sets the unix terminal mode (attributes).
 *
 * @param p0 the file descriptor
 * @param p1 the terminal mode
 */
void startup_unix_terminal_mode_set(void* p0, void* p1) {

    if (p1 != *NULL_POINTER_STATE_CYBOI_MODEL) {

        struct termios* m = (struct termios*) p1;

        if (p0 != *NULL_POINTER_STATE_CYBOI_MODEL) {

            int* f = (int*) p0;

            log_message_terminated((void*) DEBUG_LEVEL_LOG_CYBOI_MODEL, (void*) L"Startup unix terminal mode set.");

            //
            // Initialise error number.
            //
            // It is a global variable / function and other operations
            // may have set some value that is not wanted here.
            //
            // CAUTION! Initialise the error number BEFORE calling
            // the function that might cause an error.
            //
            copy_integer((void*) &errno, (void*) NUMBER_0_INTEGER_STATE_CYBOI_MODEL);

            //
            // Set terminal mode (attributes).
            //
            // The second argument specifies how to deal with
            // input and output already queued.
            // It can be one of the following values:
            // TCSANOW - Make the change immediately.
            // TCSADRAIN - Make the change after waiting until all queued output has been written. You should usually use this option when changing parameters that affect output.
            // TCSAFLUSH - This is like TCSADRAIN, but also discards any queued input.
            // TCSASOFT - This is a flag bit that you can add to any of the above alternatives.
            //            Its meaning is to inhibit alteration of the state of the terminal hardware.
            //            It is a BSD extension; it is only supported on BSD systems and the GNU system.
            //            Using TCSASOFT is exactly the same as setting the CIGNORE bit in the c_cflag member of the structure termios-p points to.
            //
            int e = tcsetattr(*f, TCSANOW, m);

            if (e < *NUMBER_0_INTEGER_STATE_CYBOI_MODEL) {

                fwprintf(stdout, L"Could not startup unix terminal mode set. An error occured within tcsetattr.\n");

                if (errno == EBADF) {

                    log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. The filedes argument is not a valid file descriptor.");
                    fwprintf(stdout, L"Could not startup unix terminal mode set. The filedes argument is not a valid file descriptor. EBADF errno: %i\n", errno);

                } else if (errno == ENOTTY) {

                    log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. The filedes is not associated with a unix terminal.");
                    fwprintf(stdout, L"Could not startup unix terminal mode set. The filedes is not associated with a unix terminal. ENOTTY errno: %i\n", errno);

                } else if (errno == EINVAL) {

                    log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. Either the value of the second argument is not valid, or there is something wrong with the data in the third argument.");
                    fwprintf(stdout, L"Could not startup unix terminal mode set. Either the value of the second argument is not valid, or there is something wrong with the data in the third argument. EINVAL errno: %i\n", errno);

                } else {

                    log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. An unknown error occured.");
                    fwprintf(stdout, L"Could not startup unix terminal mode set. An unknown error occured. UNKNOWN errno: %i\n", errno);
                }
            }

        } else {

            log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. The file descriptor is null.");
        }

    } else {

        log_message_terminated((void*) ERROR_LEVEL_LOG_CYBOI_MODEL, (void*) L"Could not startup unix terminal mode set. The original terminal mode is null.");
    }
}

/* SET_MODE_UNIX_TERMINAL_STARTER_SOURCE */
#endif
